package coproblem;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import org.apache.commons.lang.Validate;

//import com.sun.org.apache.bcel.internal.generic.INSTANCEOF;

import boundedmaxsum.BoundedDiscreteMaxSumController;
import boundedmaxsum.BoundedInternalFunction;
import boundedmaxsum.Distribution;
import boundedmaxsum.TwoVariablesRandomPayoffFunction;

import maxsum.discrete.DiscreteInternalFunction;
import maxsum.discrete.DiscreteVariableState;
import maxsum.Function;
import maxsum.Variable;
import maxsum.discrete.DiscreteInternalVariable;
import maxsum.discrete.DiscreteVariable;
import maxsum.discrete.DiscreteVariableDomain;
import maxsum.discrete.DiscreteVariableDomainImpl;
import maxsum.discrete.OptimisedDiscreteMarginalMaximisation;
import maxsum.discrete.VariableJointState;
import example.Color;
import example.ColorDomain;

/**
 * descrie a COP formulation of a factor graph
 * @author sandrof
 *
 */
public class COProblem {
	private static final double MULT_CONST = 1E4;
	protected Set<? extends DiscreteInternalVariable> vars;
	protected Set<? extends DiscreteInternalFunction> funs;
	
	public COProblem(Set<? extends DiscreteInternalVariable> vs, Set<? extends DiscreteInternalFunction> fs) {
		vars = vs;
		funs = fs;
	}
	
	public Set<? extends DiscreteInternalFunction> getFuns() {
		return funs;
	}
	
	public Set<? extends DiscreteInternalVariable> getVars() {
		return vars;
	}

	/**
	 * saves the COP to a file using the ADOPT format
	 * 
	 * @param filename the name of the output file
	 */
	public void save(String filename) {
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename));
			out.write("AGENT 1 \n");
			
			for (DiscreteVariable v : vars) {
				int name = Integer.valueOf(v.getName());
				int size = v.getDomainSize();
				out.write("VARIABLE "+name+" 1 "+size+"\n");
			}
			
			for (DiscreteInternalFunction f : funs) {
				if (f instanceof TwoVariablesRandomPayoffFunction) {
					TwoVariablesRandomPayoffFunction rf = (TwoVariablesRandomPayoffFunction) f;
					List dv = new LinkedList(rf.getVariableDependencies());
					DiscreteVariable var0 = (DiscreteVariable) dv.get(0);
					DiscreteVariable var1 = (DiscreteVariable) dv.get(1);
					int id0 = Integer.valueOf(var0.getName());
					int id1 = Integer.valueOf(var1.getName());
					
					out.write("CONSTRAINT "+id0+" "+id1+"\n");
					
					for (Color varState0 : ((DiscreteVariableDomainImpl<Color>) var0.getDomain()).getStates()) {
						for (Color varState1 : ((DiscreteVariableDomainImpl<Color>) var1.getDomain()).getStates()) {
							Map<DiscreteVariable<?>, DiscreteVariableState> currentJointStateMap = new HashMap<DiscreteVariable<?>, DiscreteVariableState>();
							Map<DiscreteVariable<?>, DiscreteVariableState> reverseJointStateMap = new HashMap<DiscreteVariable<?>, DiscreteVariableState>();
							currentJointStateMap.put(var0, varState0);
							currentJointStateMap.put(var1, varState1);
							VariableJointState currentJointState = new VariableJointState(currentJointStateMap);
							double value = rf.evaluate(currentJointState);
							out.write("F "+varState0.getName()+" "+varState1.getName()+" "+ (int)(( TwoVariablesRandomPayoffFunction.MAX_VALUE - value)*MULT_CONST)+"\n");	
						}
					}	
					
				} else {
					System.out.println("cannot deal with this type of constraint "+f.getClass());
					System.exit(0);
				}
			}
			
			out.close();
			
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
	
	
	public double getAlpha(){
		double max = -1.;
		for (DiscreteInternalFunction f : funs) {
			if (f instanceof TwoVariablesRandomPayoffFunction) {
				TwoVariablesRandomPayoffFunction tvf = (TwoVariablesRandomPayoffFunction) f;
				double curr = tvf.getMaximumBound()/tvf.getMinimumBound();
				if (curr>max){
					max=curr;
				}
			}
		}
		return max;
	}
	
	public double getBeta(){
		return 1/getAlpha();
	}
	
	public static COProblem load(String fname){
		try {
			Set vs = new HashSet();
			Set fs = new HashSet();

			Map<Integer, DiscreteVariable> mapVidToV = new HashMap<Integer, DiscreteVariable>();

			BufferedReader in = new BufferedReader(new InputStreamReader(
					new FileInputStream(fname)));
			String line = in.readLine();
			while (line != null) {
				StringTokenizer t = new StringTokenizer(line);
				String ct = t.nextToken();
				if (ct.equals("VARIABLE")) {
					int vid = new Integer(t.nextToken()).intValue();
					t.nextToken();	//skipping agent id	
					int numberOfColors = new Integer(t.nextToken()).intValue();
					DiscreteVariableDomain<Color> domain = new ColorDomain(numberOfColors);

					DiscreteInternalVariable<Color> cv = new DiscreteInternalVariable<Color>(
							""+vid, domain);
					cv.setOwningAgentIdentifier("void");
					vs.add(cv);		
					mapVidToV.put(vid,cv);
				}
				if (ct.equals("CONSTRAINT")) {
					int vid1 = new Integer(t.nextToken()).intValue();
					int vid2 = new Integer(t.nextToken()).intValue();
					DiscreteVariable var1 = mapVidToV.get(vid1);
					DiscreteVariable var2 = mapVidToV.get(vid2);
					
					TwoVariablesRandomPayoffFunction rf = new TwoVariablesRandomPayoffFunction(
							"r"+vid1+"-"+vid2, var1, var2,Distribution.UNIFORM);
					rf.setOwningAgentIdentifier("void");
					Map<VariableJointState, Double> payoffMatrix = new HashMap<VariableJointState, Double>();
					int jointStatenumber = var1.getDomainSize() * var2.getDomainSize();
					
					for (int i = 0; i < jointStatenumber; i++) {
						line = in.readLine();
						StringTokenizer it = new StringTokenizer(line);
						Validate.isTrue(it.nextToken().equals("F"));
						int varState1 = new Integer(it.nextToken()).intValue();
						int varState2 = new Integer(it.nextToken()).intValue();
						int currentJointStateValue = new Integer(it.nextToken()).intValue();
						double currentJointStateValueD = TwoVariablesRandomPayoffFunction.MAX_VALUE - (currentJointStateValue/MULT_CONST);
						
						HashMap<DiscreteVariable<Color>, DiscreteVariableState> currentJointStateMap = new HashMap<DiscreteVariable<Color>, DiscreteVariableState>();
						Color c1 = (Color) ((DiscreteVariableDomainImpl<Color>) var1.getDomain()).getStates()
						.toArray()[varState1];
						Color c2 = (Color) ((DiscreteVariableDomainImpl<Color>) var2.getDomain()).getStates()
						.toArray()[varState2];
						currentJointStateMap.put(var1, c1);
						currentJointStateMap.put(var2, c2);
						
						Map<DiscreteVariable<?>, DiscreteVariableState> reverseJointStateMap = new HashMap<DiscreteVariable<?>, DiscreteVariableState>();
						reverseJointStateMap .put(var2, c2);
						reverseJointStateMap.put(var1, c1);

						
						VariableJointState vjs = new VariableJointState(currentJointStateMap);		
						VariableJointState rvjs = new VariableJointState(reverseJointStateMap);		

						payoffMatrix.put(vjs, (double) currentJointStateValueD);
						payoffMatrix.put(rvjs, (double) currentJointStateValueD);

					}	
					rf.setPayoffMatrix(payoffMatrix);

					rf.addVariableDependency(var1);
					rf.addVariableDependency(var2);
					
					
					Validate.isTrue(fs.add(rf));
					
				}
				line = in.readLine();
				//System.out.println(line);
			}	
			return new COProblem(vs,fs); 
		} catch (IOException e) {
			System.out.println("Exception in file reading" + e);
			return null;
		}
					
	}

	/*
	 * example of ADOPT format

			AGENT 1
			VARIABLE 0 1 2
			VARIABLE 1 1 2
			VARIABLE 2 1 2 
			CONSTRAINT 0 1
			F 0 0 1
			F 0 1 2
			F 1 0 3
			F 1 1 4
			CONSTRAINT 0 2
			F 0 0 3
			F 0 1 2
			F 1 0 4
			F 1 1 1
			CONSTRAINT 1 2
			F 0 0 3
			F 0 1 4
			F 1 0 1
			F 1 1 2

	 *
	 */
	
	
	public static void main(String[] args) {
		
		int numberOfColors = 2;
		DiscreteVariableDomain<Color> domain = new ColorDomain(numberOfColors);

		DiscreteInternalVariable<Color> v1 = new DiscreteInternalVariable<Color>(
				"0", domain);
		DiscreteInternalVariable<Color> v2 = new DiscreteInternalVariable<Color>(
				"1", domain);
		DiscreteInternalVariable<Color> v3 = new DiscreteInternalVariable<Color>(
				"2", domain);

		TwoVariablesRandomPayoffFunction rf1 = new TwoVariablesRandomPayoffFunction(
				"rf1", v1, v2,Distribution.UNIFORM);

		TwoVariablesRandomPayoffFunction rf2 = new TwoVariablesRandomPayoffFunction(
				"rf2", v2, v3,Distribution.UNIFORM);

		TwoVariablesRandomPayoffFunction rf3 = new TwoVariablesRandomPayoffFunction(
				"rf3", v1, v3,Distribution.UNIFORM);

		rf1.addVariableDependency(v1);
		rf1.addVariableDependency(v2);
		rf2.addVariableDependency(v2);
		rf2.addVariableDependency(v3);
		rf3.addVariableDependency(v1);
		rf3.addVariableDependency(v3);
		
		System.out.println(rf1.getPayoffMatrix());
		System.out.println(rf2.getPayoffMatrix());
		System.out.println(rf3.getPayoffMatrix());
		
		BoundedDiscreteMaxSumController controller = new BoundedDiscreteMaxSumController("controller",new OptimisedDiscreteMarginalMaximisation());

		controller.addInternalFunction(rf1);
		controller.addInternalFunction(rf2);
		controller.addInternalFunction(rf3);

		controller.addInternalVariable(v1);
		controller.addInternalVariable(v2);
		controller.addInternalVariable(v3);
		
		COProblem p = new COProblem(controller.getInternalVariables(),controller.getInternalFunctions());
		p.save("testCOPSave.COP");
		
		controller.synchroniseInitFactorGraph();
		System.out.println(controller.getOptimalValue());
		System.out.println(controller.getOptimalState());
		
		
		COProblem lp = COProblem.load("testCOPSave.COP");
		System.out.println(lp.funs);
		for (Function f : lp.funs) {
			System.out.println(f + " dep "+f.getDependencies());
			System.out.println(f + " pm "+((TwoVariablesRandomPayoffFunction) f).getPayoffMatrix());
			
		}
		
		System.out.println(lp.vars);
		for (Variable v : lp.vars) {
			System.out.println(v + " dep "+v.getDependencies());
		}
		
		System.out.println("beta = "+lp.getBeta());
		System.out.println("alpha = "+lp.getAlpha());
		
		
	}
	
}
